# -*- coding: utf-8 -*-
"""
Functions for MCS_052_main.py

@author: Silvan Schwebke
"""

from numpy import *
import scipy.spatial as spt
import heapq
from bootstrap import ci
from random import SystemRandom
import matplotlib.pyplot as plt
import cmath
import copy

def build_listoftuples(resistance0):
    # Create list of 3-tuples (i,j,R)
    # for usage in resistance.calculate_resistance()
    res_list = []
    length = len(resistance0)
    for i in range(length):
        res_list.append((int(resistance0[i,0]),int(resistance0[i,1]),resistance0[i,2]))
    return res_list

def resistance_total(dist, a, b, c):
   # Calculate total resistance between two particles
   r_cluster = 10.0
   r_matrix = a * dist
   r_tunnel = b * exp(c*dist)
   r_total = r_cluster + r_matrix * r_tunnel / ( r_matrix + r_tunnel )
   return r_total

def create_grid(n,ax,ay,az,d,sigma):
    rng = SystemRandom()
    random.seed(rng.randint(0, 4294967295))
    coordx = random.random(n) * ax
    coordy = random.random(n) * ay
    if az>0.0:
        coordz = random.random(n) * az
    else:
        coordz = zeros(n)   # 2-dimensional case
    diameters = sigma * random.standard_normal(n) + d
    #diameters = ones((n)) * 5.0
    return (coordx, coordy, coordz, diameters)

def create_grid_hexagonal(n,ax,ay,az,d,sigma,h,phi):
    coordx = []
    coordy = []
    coordz = []
    cosphi = cos(phi*pi/180.0)
    sinphi = sin(phi*pi/180.0)
    # Find spacing parameter
    hx = h * sin(30.*pi/180.0)
    hy = h * cos(30.*pi/180.0)
    hz = h * sqrt(6.)/3.
    # number of layers in z
    nz = int( 1+floor(az/h) )
    for iz in range(nz):
        # Start at the center
        n_u = -int( sinphi * ax/hy )
        n_o = int( ceil( cosphi * ay/hy ) )
        for i in range(n_u,n_o):
            if i%2:
                x0 = cosphi * hx - sinphi * hy * float(i)
                y0 = sinphi * hx + cosphi * hy * float(i)
            else:
                x0 = -sinphi * float(i) * hy
                y0 = cosphi * float(i) * hy
            x = x0
            y = y0
            j = 0
            while x<ax and y<ay:
                if iz%2==0:
                    x = x0 + float(j) * h * cosphi
                    y = y0 + float(j) * h * sinphi
                else:
                    x = x0 + float(j) * h * cosphi + h/sqrt(3.) * sinphi
                    y = y0 + float(j) * h * sinphi - h/sqrt(3.) * cosphi
                j += 1
                if x>=0.0 and x<=ax and y>=0.0 and y<=ay:
                    coordx = hstack((coordx,x))
                    coordy = hstack((coordy,y))
                    coordz = hstack((coordz,iz*hz))
    n = len(coordx)
    diameters = sigma * random.standard_normal(n) + d
#    diameters = random.lognormal(log(d),sigma,n)
    # Make sure all diameters are >=0
    for i in range(len(diameters)):
        if diameters[i]< 0.0:
            diameters[i] = abs(diameters[i]) 
    return (coordx, coordy, coordz, diameters, n)

def MCS_find_neighbors(n, n_neighbors, coordx, coordy, coordz, diameters, ax, ay, az, xdeform=1., ydeform=1., zdeform=1., E_ratio=0.0, separation_list_only=False):
#    if strain_proportional==False:
#        coordx = coordx*xdeform; coordy = coordy*ydeform; coordz = coordz*zdeform
#        ax = ax*xdeform; ay = ay*ydeform; az*xdeform
    clusters = vstack([coordx,coordy,coordz]).T
    tree = spt.KDTree(clusters)
    tree_dist = zeros((n, n_neighbors+1))
    tree_ind = zeros((n, n_neighbors+1),dtype=int32)
    
    for i in range(n):
        # der erste Nachbar (row 0) ist immer der Cluster selbst - später nicht berücksichtigen!
        tree_dist[i,:], tree_ind[i,:] = tree.query(clusters[i], k=n_neighbors+1)
    
    list_clusters = vstack((zeros(n_neighbors),tree_ind[0,1:])).T
    list_dist = tree_dist[0,1:]
    for i in range(1,n):
        list_clusters = concatenate([list_clusters,vstack([i*ones(n_neighbors),tree_ind[i,1:]]).T])
        list_dist = hstack([list_dist,tree_dist[i,1:]])
    
    kanten = vstack((list_clusters.T,list_dist)).T
    
    #n_kanten = len(kanten)
    
    ### Add 2 points (zero diameter) which represent the contacts at the outer surfaces
    ### Since they are y-z planes, only evaluate x distances to particles
    # First define number of neighbor particles to consider for the electrodes,
    # depending on the aspect ratio of the total volume.
    if az == 0.0:
        n_endcontacts = int(ceil(sqrt(n*ay/ax)))
    else:
        n_endcontacts = int(ceil( ( n / (ax*ay*az/((min(ax,ay,az))**3.)) )**(2./3.) * max(az/ay,ay/az) ))
    n_endcontacts = max(n_endcontacts, n_neighbors)
    #print n_endcontacts
    heap_smallest = heapq.nsmallest(n_endcontacts, enumerate(coordx), key=lambda x: x[1])
    heap_largest = heapq.nlargest(n_endcontacts, enumerate(coordx), key=lambda x: x[1])
    newpoint1 = n
    newpoint2 = n+1
    for i in range(n_endcontacts):
        kanten = vstack((kanten,(newpoint1,heap_smallest[i][0],heap_smallest[i][1])))
    for i in range(n_endcontacts):
        kanten = vstack((kanten,(newpoint2,heap_largest[i][0],ax-heap_largest[i][1])))
    for i in range(n_endcontacts):
        kanten = vstack((kanten,(heap_smallest[i][0],newpoint1,heap_smallest[i][1])))
    for i in range(n_endcontacts):
        kanten = vstack((kanten,(heap_largest[i][0],newpoint2,ax-heap_largest[i][1])))
    
    diam_all = hstack((diameters,0.,0.))
    
    # Fix distances: subtract radii
    for i in range(shape(kanten)[0]):
        kanten[i,2] = max(0.0, kanten[i,2] - 0.5* (diam_all[kanten[i,0]]+diam_all[kanten[i,1]]))
    
    # Strain distances
    for i in range(len(kanten)):
        if kanten[i,2]!=0.0:
            s = kanten[i,2]
            d = 0.5* (diam_all[kanten[i,0]]+diam_all[kanten[i,1]])
            if kanten[i,0]<len(coordx) and kanten[i,1]<len(coordx):
                sx = abs( coordx[kanten[i,0]] - coordx[kanten[i,1]] )
                sy = abs( coordy[kanten[i,0]] - coordy[kanten[i,1]] )
            else:
                sx = s
                sy = 0.
            if sx!=0. or sy!=0.:
                epsilon_s = sqrt( ((sx*xdeform)**2. + (sy*ydeform)**2.) / (sx**2. + sy**2.) ) - 1
            else:
                epsilon_s = 0.0
            if s!=0.0:
                kanten[i,2] = s * ( 1 + epsilon_s * (s+d)/(s+d*E_ratio) )

    list_dist = copy.deepcopy(kanten[:,2])

    if separation_list_only:
        return list_dist
    
    return (clusters, kanten, newpoint1, newpoint2, tree_dist, tree_ind, list_dist)

def MCS_fix_diameters(tree_dist, tree_ind, diameters, n, n_neighbors, two_dimensions):
    # Fix diameters of intersecting particles
    tree_dist_temp = zeros((n, n_neighbors))
    tree_ind_temp = zeros((n, n_neighbors),dtype=int32)
    # Drop first column, it contains the distance of the node to itself
    for ii in range(shape(tree_dist_temp)[0]):
        tree_dist_temp[ii,:] = tree_dist[ii,-n_neighbors:] - diameters[ii]/2. - diameters[tree_ind[ii,-n_neighbors:]]/2.
        tree_ind_temp[ii,:] = tree_ind[ii,-n_neighbors:]
    
    diameters_fixed = zeros(n)
    for jj in range(shape(tree_dist_temp)[0]):
        if two_dimensions==True:
            area_fix = 0.0
            for kk in range(shape(tree_dist_temp)[1]):
                if tree_dist_temp[jj,kk] < 0.0:
                    d = tree_dist[jj,kk+1] # d is the center to center distance
                    R = diameters[jj]/2.0
                    r = diameters[tree_ind_temp[jj,kk]]/2.0
                    # if d<=abs(R-r), then one circle is completely within the other
                    # -> formula does not apply anymore intersection area is equal to smaller circle area
                    if d>abs(R-r):
                        area_intersect = r*r * arccos( (d*d + r*r - R*R) / (2.*d*r) ) + R*R * arccos( (d*d + R*R - r*r) / (2.*d*R) ) - 0.5 * sqrt( (-d+r+R) * (d+r-R) * (d-r+R) * (d+r+R) )
                    else:
                        area_intersect = pi * (min((R,r)))**2.
                    area_R = pi * R**2.
                    if area_intersect > 0.:
                        area_fix += 1. / ( 1./(0.5*area_intersect) - 1./area_R )
                    # otherwise area_fix stays the same
            diameters_fixed[jj] = sqrt( (diameters[jj])**2. + 4.*area_fix/pi )
        else:
            vol_fix = 0.0
            for kk in range(shape(tree_dist_temp)[1]):
                if tree_dist_temp[jj,kk] < 0.0:
                    d = tree_dist[jj,kk+1] # d is the center to center distance
                    R = diameters[jj]/2.0
                    r = diameters[tree_ind_temp[jj,kk]]/2.0
                    # if d<=abs(R-r), then one circle is completely within the other
                    # -> formula does not apply anymore intersection area is equal to smaller circle area
                    if d>abs(R-r):
                        vol_intersect = 1./(12.*d) * pi * (R+r-d)**2. * (d*d + 2.*d*r - 3.*r*r + 2.*d*R + 6.*r*R - 3.*R*R)
                    else:
                        vol_intersect = 4./3.*pi * (min((R,r)))**3.
                    vol_R = 4./3.*pi*R**3.
                    if vol_intersect > 0.:
                        vol_fix += 1. / ( 1./(0.5*vol_intersect) - 1./vol_R )
                    # otherwise vol_fix stays the same
            diameters_fixed[jj] = 2. * ( (diameters[jj]/2.)**3. + 3.*vol_fix/(4.*pi) )**(1./3.)
    return diameters_fixed

def calc_uncertainty_k(path_resistance,uncertainty_resistance,kLkT, deformation):
    (R0, RL, RT) = path_resistance
    (u_R0, u_RL, u_RT) = uncertainty_resistance
    (kL, kT) = kLkT
    epsilon = deformation - 1.0
    u_kL = abs(RL/(R0*R0) / epsilon) * u_R0 + abs(1./R0 / epsilon) * u_RL
    u_kT = abs(RT/(R0*R0) / epsilon) * u_R0 + abs(1./R0 / epsilon) * u_RT
    return (u_kL, u_kT)

def calc_accuracy(array_kLkT, i, error_goal_k, thisalpha=0.05, thisoutput='errorbar', print_result=True, samples=5000, max_only=False):
    if ndim(array_kLkT)==1:
        error_kL = ci(array_kLkT[:i], mean, alpha=thisalpha, output=thisoutput, n_samples=samples)
        error_rel_kL = error_kL/mean(array_kLkT[:i])
        error_max_rel = max(error_rel_kL)
        error_max_abs = max(error_kL)
        error_coefficient = error_max_rel * sqrt(i)
        n_loops_prediction = int(ceil( (error_coefficient/error_goal_k)**2. ))
        if max_only:
            return (error_max_rel, error_max_abs)
        else:
            return ( (float(error_kL[0]),float(error_kL[1])), (float(error_rel_kL[0]),float(error_rel_kL[1])), error_max_rel)
    error_kL = ci(array_kLkT[:i,0], mean, alpha=thisalpha, output=thisoutput, n_samples=samples)
    error_rel_kL = error_kL/mean(array_kLkT[:i,0])
    error_kT = ci(array_kLkT[:i,1], mean, alpha=thisalpha, output=thisoutput, n_samples=samples)
    error_rel_kT = error_kT/mean(array_kLkT[:i,1])
    error_max_rel = max( vstack( (error_rel_kL,error_rel_kT) ) )
    error_max_abs = max( vstack( (error_kL,error_kT) ) )
    error_coefficient = error_max_rel * sqrt(i)
    n_loops_prediction = int(ceil( (error_coefficient/error_goal_k)**2. ))
    if print_result:
        print 'rel. uncert. (95% interval) kL: +{:.4f}, -{:.4f}; kT: +{:.4f}, -{:.4f}.\nEstimated no. of loops for e<={:.4f}: {}'.format(float(error_rel_kL[1]),float(error_rel_kL[0]),float(error_rel_kT[1]),float(error_rel_kT[0]),error_goal_k,n_loops_prediction)
    if max_only:
        return (error_max_rel, error_max_abs)
    else:
        return ( (float(error_kL[0]),float(error_kL[1])), (float(error_rel_kL[0]),float(error_rel_kL[1])), (float(error_kT[0]),float(error_kT[1])), (float(error_rel_kT[0]),float(error_rel_kT[1])), error_max_rel)
    
def calc_separation_distribution(kanten,numBins):
    fig = plt.figure()
    ax = fig.add_subplot(111)
    ax.hist(kanten,numBins,color='green',alpha=0.8)
    plt.show()
    return
    
def mole_to_volume_fraction(molefraction_metal, metal, matrix):
    # Convert mole fraction to volume fraction, to be used in the further calculation.
    # Dictionary of material data: molar mass a in u, density rho in kg/m³
    material_dict_a_rho = {'Al2O3': (102.0, 4000.),
                          'Pt': (195.1, 21450.),
                          'BN': (24.82, 2250.), # hBN: 2250 kg/m³, cBN: 3450 kg/m³
                          'Ni': (58.69, 8908.),
                          'C': (12.01, 2260.),  # Graphite: 2260 kg/m³, Diamond: 3515 kg/m³
                         }
    (a1, rho1) = material_dict_a_rho[metal]
    (a2, rho2) = material_dict_a_rho[matrix]
    volfrac = zeros(len(molefraction_metal))
    for i in range(len(molefraction_metal)):
        x1 = molefraction_metal[i]
        x2 = 1. - x1
        volfrac[i] = 1. / ( 1. + (x2*a2*rho1)/(x1*a1*rho2) )
    return volfrac

def elastic_modulus(metal, matrix):
    # Material data: elastic modulus in GPa
    material_dict_E = {'Al2O3': 380.,
                      'Pt': 168.,
                      'BN': 80.,    # hBN: 33 GPa, cBN: 415 (Thin films [Hahn97])
                      'Ni': 200.,
                      'C': 80.,    # Graphite: 4.8 GPa, Diamond: 950 GPa; a-C:H: 25 GPa, ta-C: 200... 820 GPa (Thin films [Hetzner14])
                     }
    return (material_dict_E[metal], material_dict_E[matrix])
    
    
